package muesli

import "fmt"

/* Run time values */
type Value interface {
	String() string
	Type() TypeValue
}

type IntValue int64
type FloatValue float64
type StringValue string
type BoolValue bool
type WordValue string
type TypeValue string
type ErrorValue struct {
	error
}
type EmptyValue struct {
}
type ListValue struct {
	List []Value
}

const (
	TrueValue       = BoolValue(true)
	FalseValue      = BoolValue(false)
	IntTypeValue    = TypeValue("Int")
	FloatTypeValue  = TypeValue("Float")
	StringTypeValue = TypeValue("String")
	BoolTypeValue   = TypeValue("Bool")
	WordTypeValue   = TypeValue("Word")
	ErrorTypeValue  = TypeValue("Error")
	TypeTypeValue   = TypeValue("Type")
	EmptyTypeValue  = TypeValue("Empty")
	ListTypeValue   = TypeValue("List")
	AnyTypeValue    = TypeValue("Any")
)

var NilValue = Value(nil)

func (val IntValue) String() string {
	return fmt.Sprintf("%d", int64(val))
}

func (val FloatValue) String() string {
	return fmt.Sprintf("%f", float64(val))
}

func (val BoolValue) String() string {
	if bool(val) {
		return "true"
	} else {
		return "false"
	}
}

func (val StringValue) String() string {
	return string(val)
}

func (val WordValue) String() string {
	return string(val)
}

func (val TypeValue) String() string {
	return string(val)
}

func (val ErrorValue) String() string {
	return fmt.Sprintf("%s", val.Error())
}

func (val EmptyValue) String() string {
	return "<empty>"
}

func (val ListValue) String() string {
	res := "["
	sep := ""
	for _, elt := range val.List {
		res = res + sep + elt.String()
		sep = ", "
	}
	res += "]"
	return res
}

func (v IntValue) Type() TypeValue    { return IntTypeValue }
func (v FloatValue) Type() TypeValue  { return FloatTypeValue }
func (v StringValue) Type() TypeValue { return StringTypeValue }
func (v BoolValue) Type() TypeValue   { return BoolTypeValue }
func (v WordValue) Type() TypeValue   { return WordTypeValue }
func (v TypeValue) Type() TypeValue   { return TypeTypeValue }
func (v ErrorValue) Type() TypeValue  { return ErrorTypeValue }
func (v EmptyValue) Type() TypeValue  { return EmptyTypeValue }
func (v ListValue) Type() TypeValue   { return ListTypeValue }

func NewErrorValuef(format string, args ...interface{}) ErrorValue {
	err := fmt.Errorf(format, args...)
	return ErrorValue{err}
}

func NewListValue(elements ...Value) ListValue {
	return ListValue{elements}
}

func (list *ListValue) Append(elements ...Value) {
	list.List = append(list.List, elements...)
}
